import pytest
from collections import Counter

from small_language_model import SmallLanguageModel


def test_initializes_with_empty_dict() -> None:
    slm = SmallLanguageModel()

    assert len(slm.char_map) == 0


@pytest.fixture
def initialized_slm() -> SmallLanguageModel:
    slm = SmallLanguageModel()
    slm.train("Hello")

    return slm


def test_train(initialized_slm):
    correct_H = Counter({"e": 1})
    assert initialized_slm.char_map["H"] == correct_H

    correct_e = Counter({"l": 1})
    assert initialized_slm.char_map["e"] == correct_e

    correct_l = Counter({"l": 1, "o": 1})
    assert initialized_slm.char_map["l"] == correct_l

    # There should not be any extra letters in the dict.
    # Note the final "o" in "Hello" should not be in the dict.
    assert set(initialized_slm.char_map.keys()) == set("Hell")


def test_predict(initialized_slm: SmallLanguageModel) -> None:
    assert initialized_slm.predict_next("H") == "e"

    assert initialized_slm.predict_next("e") == "l"


# This mark + the function below are intended to be equal to the test_predict method above
@pytest.mark.parametrize("input, expected_prediction", [("H", "e"), ("e", "l")])
def test_predicted_expected(
    initialized_slm: SmallLanguageModel, input: str, expected_prediction: str
) -> None:
    assert initialized_slm.predict_next(input) == expected_prediction


def test_edge_cases(initialized_slm: SmallLanguageModel) -> None:
    assert initialized_slm.predict_next("o") is None
    assert initialized_slm.predict_next("z") is None



from collections import defaultdict, Counter
import random
from typing import DefaultDict, Counter as TypingCounter, List, Tuple


class SmallLanguageModel:
    def __init__(self) -> None:
        self.char_map: DefaultDict[str, TypingCounter[str]] = defaultdict(Counter)

    def train(self, text: str) -> None:
        if len(text) < 2:
            raise Exception("Text is too short")

        for i in range(len(text) - 1):
            current_char = text[i]
            next_char = text[i + 1]
            self.char_map[current_char][next_char] += 1

    def predict_next(self, current_char: str) -> str | None:
        if current_char not in self.char_map:
            return None
        next_chars, weights = zip(*self.char_map[current_char].items())
        if next_chars:
            chosen_char: str | None = random.choices(next_chars, weights=weights)[0]
        else:
            chosen_char = None
        return chosen_char

    def get_character_training_frequency(self) -> List[Tuple[str, int]]:
        """
        Returns the number of times each character appears in the training data.

        Returns a list of tuples. The first character in the tuple is the character.
        The second character is the number of times it appears. The list is sorted
        in descending order.
        """
        # The total frequency of a letter is the sum of all the letters which come after it
        num_occurences: dict[str, int] = {}
        for k, v in self.char_map.items():
            num_occurences[k] = sum(v.values())

        # Now sort and return the frequency data
        return sorted(num_occurences.items(), key=lambda x: x[1], reverse=True)


# Example usage
if __name__ == "__main__":
    slm = SmallLanguageModel()
    slm.train("hello")
    print(slm.char_map)
    print(slm.predict_next("l"))
    print(slm.predict_next("l"))
    print(slm.get_character_training_frequency())
